{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "# Custom module example\n\nThis tutorial shows how to support a custom module in a simple fashion.\nWe focus on `BackPACK's first-order extensions <https://docs.backpack.pt/en/master/extensions.html#first-order-extensions>`_.\nThey don't backpropagate additional information and thus require less functionality be\nimplemented.\n\nLet's get the imports out of our way.\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import torch\n\nfrom backpack import backpack, extend\nfrom backpack.extensions import BatchGrad\nfrom backpack.extensions.firstorder.base import FirstOrderModuleExtension\n\n# make deterministic\ntorch.manual_seed(0)\n\ndevice = torch.device(\"cuda\" if torch.cuda.is_available() else \"cpu\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Custom PyTorch module\nIn this example, we consider extending our own, very simplistic, layer.\nIt scales the input by a scalar ``weight`` in a forward pass. Here is the\nlayer class (see https://pytorch.org/docs/stable/notes/extending.html).\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class ScaleModule(torch.nn.Module):\n    \"\"\"Defines the module.\"\"\"\n\n    def __init__(self, weight=2.0):\n        \"\"\"Store scalar weight.\n\n        Args:\n            weight(float, optional): Initial value for weight. Defaults to 2.0.\n        \"\"\"\n        super(ScaleModule, self).__init__()\n\n        self.weight = torch.nn.Parameter(torch.tensor([weight]))\n\n    def forward(self, input):\n        \"\"\"Defines forward pass.\n\n        Args:\n            input(torch.Tensor): input\n\n        Returns:\n            torch.Tensor: product of input and weight\n        \"\"\"\n        return input * self.weight"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "You don't necessarily need to write a custom layer. Any PyTorch layer can be extended\nas described (it should be a :py:class:`torch.nn.Module <torch.nn.Module>`'s because\nBackPACK uses module hooks).\n\n## Custom module extension\nLet's make BackPACK support computing individual gradients for ``ScaleModule``.\nThis is done by the :py:class:`BatchGrad <backpack.extensions.BatchGrad>` extension.\nTo support the new module, we need to create a module extension that implements\nhow individual gradients are extracted with respect to ``ScaleModule``'s parameter.\n\nThe module extension must implement methods named after the parameters passed to the\nconstructor. Here it goes.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class ScaleModuleBatchGrad(FirstOrderModuleExtension):\n    \"\"\"Extract individual gradients for ``ScaleModule``.\"\"\"\n\n    def __init__(self):\n        \"\"\"Store parameters for which individual gradients should be computed.\"\"\"\n        # specify parameter names\n        super().__init__(params=[\"weight\"])\n\n    def weight(self, ext, module, g_inp, g_out, bpQuantities):\n        \"\"\"Extract individual gradients for ScaleModule's ``weight`` parameter.\n\n        Args:\n            ext(BatchGrad): extension that is used\n            module(ScaleModule): module that performed forward pass\n            g_inp(tuple[torch.Tensor]): input gradient tensors\n            g_out(tuple[torch.Tensor]): output gradient tensors\n            bpQuantities(None): additional quantities for second-order\n\n        Returns:\n            torch.Tensor: individual gradients\n        \"\"\"\n        show_useful = True\n\n        if show_useful:\n            print(\"Useful quantities:\")\n            # output is saved under field output\n            print(\"\\tmodule.output.shape:\", module.output.shape)\n            # input i is saved under field input[i]\n            print(\"\\tmodule.input0.shape:\", module.input0.shape)\n            # gradient w.r.t output\n            print(\"\\tg_out[0].shape:     \", g_out[0].shape)\n\n        # actual computation\n        return (g_out[0] * module.input0).flatten(start_dim=1).sum(axis=1).unsqueeze(-1)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Lastly, we need to register the mapping between layer (``ScaleModule``) and layer\nextension (``ScaleModuleBatchGrad``) in an instance of\n:py:class:`BatchGrad <backpack.extensions.BatchGrad>`.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "# register module-computation mapping\nextension = BatchGrad()\nextension.set_module_extension(ScaleModule, ScaleModuleBatchGrad())"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "That's it. We can now pass ``extension`` to a\n:py:class:`with backpack(...) <backpack.backpack>` context and compute individual\ngradients with respect to ``ScaleModule``'s ``weight`` parameter.\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Test custom module\nHere, we verify the custom module extension on a small net with random inputs.\nLet's create these.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "batch_size = 10\nbatch_axis = 0\ninput_size = 4\n\ninputs = torch.randn(batch_size, input_size, device=device)\ntargets = torch.randint(0, 2, (batch_size,), device=device)\n\nreduction = [\"mean\", \"sum\"][1]\nmy_module = ScaleModule().to(device)\nlossfunc = torch.nn.CrossEntropyLoss(reduction=reduction).to(device)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "<div class=\"alert alert-info\"><h4>Note</h4><p>Results of ``\"mean\"`` and ``\"sum\"`` reduction differ by a scaling factor,\n    because the information backpropagated by PyTorch is scaled. This is documented at\n    https://docs.backpack.pt/en/master/extensions.html#backpack.extensions.BatchGrad.</p></div>\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Individual gradients with PyTorch\nThe following computes individual gradients by looping over individual samples and\nstacking their gradients.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "grad_batch_autograd = []\n\nfor input_n, target_n in zip(\n    inputs.split(1, dim=batch_axis), targets.split(1, dim=batch_axis)\n):\n    loss_n = lossfunc(my_module(input_n), target_n)\n    grad_n = torch.autograd.grad(loss_n, [my_module.weight])[0]\n    grad_batch_autograd.append(grad_n)\n\ngrad_batch_autograd = torch.stack(grad_batch_autograd)\n\nprint(\"weight.shape:             \", my_module.weight.shape)\nprint(\"grad_batch_autograd.shape:\", grad_batch_autograd.shape)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Individual gradients with BackPACK\nBackPACK can compute individual gradients in a single backward pass.\nFirst, :py:func:`extend <backpack.extend>` model and loss function, then\nperform the backpropagation inside a\n:py:class:`with backpack(...) <backpack.backpack>` context.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "my_module = extend(my_module)\nlossfunc = extend(lossfunc)\n\nloss = lossfunc(my_module(inputs), targets)\n\nwith backpack(extension):\n    loss.backward()\n\ngrad_batch_backpack = my_module.weight.grad_batch\n\nprint(\"weight.shape:             \", my_module.weight.shape)\nprint(\"grad_batch_backpack.shape:\", grad_batch_backpack.shape)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Do the computation results match?\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "match = torch.allclose(grad_batch_autograd, grad_batch_backpack)\n\nprint(f\"autograd and BackPACK individual gradients match? {match}\")\n\nif not match:\n    raise AssertionError(\n        \"Individual gradients don't match:\"\n        + f\"\\n{grad_batch_autograd}\\nvs.\\n{grad_batch_backpack}\"\n    )"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.7.9"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}